from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(label=_('Documents'), name='documents')

# Document

permission_document_create = namespace.add_permission(
    label=_('Create documents'), name='document_create'
)
permission_document_edit = namespace.add_permission(
    label=_('Edit documents'), name='document_edit'
)
permission_document_properties_edit = namespace.add_permission(
    label=_('Edit document properties'), name='document_properties_edit'
)
permission_document_tools = namespace.add_permission(
    label=_('Execute document modifying tools'), name='document_tools'
)
permission_document_view = namespace.add_permission(
    label=_('View documents'), name='document_view'
)

# Document file

permission_document_file_delete = namespace.add_permission(
    label=_('Delete document files'), name='document_file_delete'
)
permission_document_file_edit = namespace.add_permission(
    label=_('Edit document files'), name='document_file_edit'
)
permission_document_file_download = namespace.add_permission(
    label=_('Download document files'), name='document_file_download'
)
permission_document_file_new = namespace.add_permission(
    label=_('Create new document files'), name='document_file_new'
)
permission_document_file_print = namespace.add_permission(
    label=_('Print document files'), name='document_file_print'
)
permission_document_file_tools = namespace.add_permission(
    label=_('Execute document file modifying tools'),
    name='document_file_tools'
)
permission_document_file_view = namespace.add_permission(
    label=_('View document files'),
    name='document_file_view'
)

# Document version

permission_document_version_create = namespace.add_permission(
    label=_('Create document versions'),
    name='document_version_create'
)
permission_document_version_delete = namespace.add_permission(
    label=_('Delete document versions'),
    name='document_version_delete'
)
permission_document_version_edit = namespace.add_permission(
    label=_('Edit document versions'),
    name='document_version_edit'
)
permission_document_version_export = namespace.add_permission(
    label=_('Export document versions'),
    name='document_version_export'
)
permission_document_version_print = namespace.add_permission(
    label=_('Print document versions'), name='document_version_print'
)
permission_document_version_view = namespace.add_permission(
    label=_('View document versions'),
    name='document_version_view'
)

# Document type

setup_namespace = PermissionNamespace(
    label=_('Document types'), name='documents_types'
)
permission_document_type_create = setup_namespace.add_permission(
    label=_('Create document types'), name='document_type_create'
)
permission_document_type_delete = setup_namespace.add_permission(
    label=_('Delete document types'), name='document_type_delete'
)
permission_document_type_edit = setup_namespace.add_permission(
    label=_('Edit document types'), name='document_type_edit'
)
permission_document_type_view = setup_namespace.add_permission(
    label=_('View document types'), name='document_type_view'
)

# Trashed document

permission_trashed_document_restore = namespace.add_permission(
    label=_('Restore trashed document'), name='document_restore'
)
permission_trashed_document_delete = namespace.add_permission(
    label=_('Delete trashed documents'), name='document_delete'
)
permission_document_trash = namespace.add_permission(
    label=_('Trash documents'), name='document_trash'
)
permission_trash_empty = namespace.add_permission(
    label=_('Empty trash'), name='document_empty_trash'
)



from django.contrib.contenttypes.models import ContentType

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.messaging.events import event_message_created
from mayan.apps.messaging.models import Message
from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.storage.events import event_download_file_created
from mayan.apps.storage.models import DownloadFile

from ..events import event_events_exported
from ..models import Notification
from ..permissions import permission_events_export, permission_events_view

from .mixins import (
    EventsExportViewTestMixin, EventTypeTestMixin, EventViewTestMixin,
    NotificationTestMixin, NotificationViewTestMixin, UserEventViewsTestMixin
)


class EventsViewTestCase(
    EventTypeTestMixin, EventViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_user()
        self.test_object = self.test_document_type

        content_type = ContentType.objects.get_for_model(model=self.test_object)

        self.view_arguments = {
            'app_label': content_type.app_label,
            'model_name': content_type.model,
            'object_id': self.test_object.pk
        }

    def create_test_event(self, **kwargs):
        self.test_action = self.test_event_type.commit(**kwargs)
        self.test_actions.append(self.test_action)

    def test_event_list_view_no_permission(self):
        self.create_test_event(target=self.test_object)

        response = self._request_test_events_list_view()

        self.assertNotContains(
            response=response, status_code=200, text=str(self.test_event_type)
        )

    def test_event_list_view_with_access(self):
        self.create_test_event(target=self.test_object)

        self.grant_access(
            obj=self.test_object, permission=permission_events_view
        )

        response = self._request_test_events_list_view()

        self.assertContains(
            response=response, status_code=200, text=str(self.test_event_type)
        )

    def test_events_for_object_view_no_permission(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_events_for_object_view()
        self.assertNotContains(
            response=response, text=str(self.test_event_type), status_code=200
        )

    def test_events_for_object_view_with_access(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        self.grant_access(
            obj=self.test_object, permission=permission_events_view
        )

        response = self._request_events_for_object_view()
        self.assertContains(
            response=response, text=str(self.test_event_type), status_code=200
        )

    def test_events_by_verb_view_no_permission(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_test_events_by_verb_view()
        self.assertContains(
            count=3,
            response=response, text=str(self.test_event_type), status_code=200
        )

    def test_events_by_verb_view_with_access(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        self.grant_access(
            obj=self.test_object, permission=permission_events_view
        )

        response = self._request_test_events_by_verb_view()
        self.assertContains(
            count=4,
            response=response, text=str(self.test_event_type), status_code=200
        )

    def test_current_user_events_view_no_permission(self):
        self.create_test_event(
            actor=self._test_case_user, action_object=self.test_object
        )

        response = self._request_test_current_user_events_view()
        self.assertNotContains(
            response=response, text=str(self.test_event_type), status_code=200
        )

    def test_current_user_events_view_with_access(self):
        self.create_test_event(
            actor=self._test_case_user, action_object=self.test_object
        )

        self.grant_access(
            obj=self.test_object, permission=permission_events_view
        )

        response = self._request_test_current_user_events_view()
        self.assertContains(
            response=response, text=str(self.test_event_type), status_code=200
        )


class EventExportViewTestCase(
    EventTypeTestMixin, EventsExportViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_user()
        self.test_object = self.test_document_type

        content_type = ContentType.objects.get_for_model(model=self.test_object)

        self.view_arguments = {
            'app_label': content_type.app_label,
            'model_name': content_type.model,
            'object_id': self.test_object.pk
        }

        ModelPermission.register(
            model=self.test_object._meta.model, permissions=(
                permission_events_export,
            )
        )

        self._clear_events()

    def create_test_event(self, **kwargs):
        self.test_action = self.test_event_type.commit(**kwargs)
        self.test_actions.append(self.test_action)

    def test_events_list_export_view_no_permission(self):
        self.create_test_event(target=self.test_object)

        response = self._request_test_events_list_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + 3 for the test
        self.assertEqual(events.count(), 4)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_events_list_export_view_with_access(self):
        self.create_test_event(target=self.test_object)

        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        response = self._request_test_events_list_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + access grant + 3 for the test
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_download_file_created.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_download_file)
        self.assertEqual(events[3].verb, event_events_exported.id)

        self.assertEqual(events[4].action_object, None)
        self.assertEqual(events[4].actor, test_message)
        self.assertEqual(events[4].target, test_message)
        self.assertEqual(events[4].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )

    def test_events_for_object_export_view_no_permission(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_events_for_object_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + 3 for the test
        self.assertEqual(events.count(), 4)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_events_for_object_export_view_with_access(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        response = self._request_events_for_object_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + access grant + 3 for the test
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_download_file_created.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_download_file)
        self.assertEqual(events[3].verb, event_events_exported.id)

        self.assertEqual(events[4].action_object, None)
        self.assertEqual(events[4].actor, test_message)
        self.assertEqual(events[4].target, test_message)
        self.assertEqual(events[4].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )

    def test_events_by_verb_export_view_no_permission(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_test_events_by_verb_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + 3 for the test
        self.assertEqual(events.count(), 4)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_events_by_verb_view_export_with_access(self):
        self.create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        response = self._request_test_events_by_verb_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + access grant + 3 for the test
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_download_file_created.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_download_file)
        self.assertEqual(events[3].verb, event_events_exported.id)

        self.assertEqual(events[4].action_object, None)
        self.assertEqual(events[4].actor, test_message)
        self.assertEqual(events[4].target, test_message)
        self.assertEqual(events[4].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )

    def test_current_user_events_export_view_no_permission(self):
        self.create_test_event(
            actor=self._test_case_user, action_object=self.test_object
        )

        response = self._request_test_current_user_events_export_view()
        self.assertNotContains(
            response=response, text=str(self.test_event_type), status_code=302
        )

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + 3 for the test
        self.assertEqual(events.count(), 4)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_current_user_events_export_view_with_access(self):
        self.create_test_event(
            actor=self._test_case_user, action_object=self.test_object
        )

        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        response = self._request_test_current_user_events_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        # Test object creation + access grant + 3 for the test
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_download_file_created.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_download_file)
        self.assertEqual(events[3].verb, event_events_exported.id)

        self.assertEqual(events[4].action_object, None)
        self.assertEqual(events[4].actor, test_message)
        self.assertEqual(events[4].target, test_message)
        self.assertEqual(events[4].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )


class NotificationViewTestCase(
    NotificationTestMixin, NotificationViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_user()

        self.test_event_type.commit(
            actor=self.test_user, action_object=self.test_document_type
        )

    def test_notification_list_view(self):
        response = self._request_test_notification_list_view()
        self.assertEqual(response.status_code, 200)

    def test_notification_mark_read_all_view(self):
        self._create_test_notification()
        notification_count = Notification.objects.get_unread().count()

        response = self._request_test_notification_mark_read_all_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Notification.objects.get_unread().count(),
            notification_count - 1
        )

    def test_notification_mark_read_view(self):
        self._create_test_notification()
        notification_count = Notification.objects.get_unread().count()

        response = self._request_test_notification_mark_read()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Notification.objects.get_unread().count(),
            notification_count - 1
        )


class UserEventViewsTestCase(UserEventViewsTestMixin, GenericViewTestCase):
    def test_user_event_type_subscription_list_view(self):
        response = self._request_test_user_event_type_subscription_list_view()
        self.assertEqual(response.status_code, 200)



import logging

from django.core.files.base import ContentFile
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _, ugettext

logger = logging.getLogger(name=__name__)


class DatabaseFileModelMixin(models.Model):
    filename = models.CharField(
        db_index=True, max_length=255, verbose_name=_('Filename')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, verbose_name=_('Date time')
    )

    class Meta:
        abstract = True

    def delete(self, *args, **kwargs):
        if self.file.name:
            self.file.storage.delete(name=self.file.name)
        return super().delete(*args, **kwargs)

    def open(self, **kwargs):
        default_kwargs = {
            'mode': self.file.file.mode,
            'name': self.file.name
        }

        default_kwargs.update(**kwargs)

        # Ensure the caller cannot specify an alternate filename.
        name = kwargs.pop('name', None)

        if name:
            logger.warning(
                'Caller tried to specify an alternate filename: %s', name
            )

        return self.file.storage.open(**default_kwargs)

    def save(self, *args, **kwargs):
        if not self.file:
            self.file = ContentFile(
                content='', name=self.filename or ugettext('Unnamed file')
            )

        self.filename = self.filename or force_text(s=self.file)
        super().save(*args, **kwargs)



from collections import Iterable
import logging

from django.apps import apps
from django.db.models.signals import post_save, pre_delete
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import ugettext as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.exceptions import ResolverPipelineError
from mayan.apps.common.utils import (
    ResolverPipelineModelAttribute, get_related_field
)
from mayan.apps.views.literals import LIST_MODE_CHOICE_LIST

from .exceptions import DynamicSearchException
from .literals import (
    DEFAULT_SCOPE_ID, DELIMITER, SCOPE_MATCH_ALL, SCOPE_MARKER,
    SCOPE_OPERATOR_CHOICES, SCOPE_OPERATOR_MARKER, SCOPE_RESULT_MAKER
)
from .settings import (
    setting_backend, setting_backend_arguments,
    setting_results_limit
)
logger = logging.getLogger(name=__name__)


class SearchBackend:
    @staticmethod
    def get_instance():
        return import_string(dotted_path=setting_backend.value)(
            **setting_backend_arguments.value
        )

    @staticmethod
    def limit_queryset(queryset):
        pk_list = queryset.values('pk')[:setting_results_limit.value]
        return queryset.filter(pk__in=pk_list)

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _search(self, global_and_search, search_model, query_string, user):
        raise NotImplementedError

    def deindex_instance(self, instance):
        raise NotImplementedError

    def index_instance(self, instance):
        raise NotImplementedError

    def search(
        self, search_model, query, user, global_and_search=False
    ):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        result = self.decode_query(
            query=query, global_and_search=global_and_search
        )

        # Recursive call to the backend's search using queries as unscoped
        # and then merge then using the corresponding operator.
        queryset = self.solve_scope(
            operators=result['operators'],
            result_scope=result['result_scope'], search_model=search_model,
            scopes=result['scopes'], user=user
        )

        if search_model.permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=search_model.permission, queryset=queryset,
                user=user
            )

        return SearchBackend.limit_queryset(queryset=queryset)

    def decode_query(self, query, global_and_search=False):
        # Clean up the query.
        # The original query is immutable, create a new
        # mutable copy.
        query.pop('_match_all', None)

        # Turn scoped query dictionary into a series of unscoped queries.
        operators = {}
        result_scope = DEFAULT_SCOPE_ID
        scope_match_all = False
        scopes = {}

        for key, value in query.items():
            scope_id = DEFAULT_SCOPE_ID

            # Check if the entry has a scope marker.
            if key.startswith(SCOPE_MARKER):
                # Remove the scope marker.
                key = key[len(SCOPE_MARKER):]

                if key.startswith(SCOPE_OPERATOR_MARKER):
                    # Check for operator.
                    # __operator_SCOPE_SCOPE=OPERATOR_SCOPE
                    key = key[len(SCOPE_OPERATOR_MARKER):]
                    operator_scopes = key[len(DELIMITER):].split(DELIMITER)
                    operator_text, result = value.split(DELIMITER)

                    operators[result] = {
                        'scopes': operator_scopes,
                        'function': SCOPE_OPERATOR_CHOICES[operator_text],
                    }
                elif key.startswith(SCOPE_RESULT_MAKER):
                    # Check for result.
                    # __result=SCOPE
                    result_scope = value
                else:
                    # Check scope match all.
                    # __SCOPE_match_all
                    if key.endswith(SCOPE_MATCH_ALL):
                        scope_id, key = key.split(DELIMITER, 1)
                        scopes.setdefault(scope_id, {})
                        scope_match_all = value.upper() == 'TRUE'
                        scopes[scope_id]['match_all'] = scope_match_all
                    else:
                        # Must be a scoped query.
                        # __SCOPE_QUERY=VALUE
                        scope_id, key = key.split(DELIMITER, 1)
                        scopes.setdefault(scope_id, {})
                        scopes[scope_id].setdefault('match_all', False)
                        scopes[scope_id].setdefault('query', {})

                        scopes[scope_id]['query'][key] = value
            else:
                scopes.setdefault(scope_id, {})
                scopes[scope_id].setdefault('match_all', scope_match_all)
                scopes[scope_id].setdefault('query', {})
                scopes[scope_id]['query'][key] = value

        return {
            'operators': operators, 'result_scope': result_scope,
            'scopes': scopes
        }

    def solve_scope(
        self, search_model, user, result_scope, scopes, operators
    ):
        if len(scopes) > 1:
            ignore_limit = True
        else:
            ignore_limit = False

        try:
            # Try scopes.
            scope = scopes[result_scope]
        except KeyError:
            try:
                # Try operators.
                operator = operators[result_scope]
            except KeyError:
                raise DynamicSearchException(
                    'Scope `{}` not found.'.format(result_scope)
                )
            else:
                result = search_model.model._meta.default_manager.none()
                for scope in operator['scopes']:
                    queryset = self.solve_scope(
                        operators=operators, result_scope=scope,
                        search_model=search_model, scopes=scopes, user=user
                    )

                    if not result:
                        result = queryset
                    else:
                        result = operator['function'](result, queryset)

                return result
        else:
            return self._search(
                global_and_search=scope['match_all'],
                ignore_limit=ignore_limit, search_model=search_model,
                query_string=scope['query'], user=user
            )


class SearchField:
    """
    Search for terms in fields that directly belong to the parent
    SearchModel.
    """
    def __init__(
        self, search_model, field, label=None, transformation_function=None
    ):
        self.search_model = search_model
        self.field = field
        self._label = label
        self.transformation_function = transformation_function

    def get_full_name(self):
        return self.field

    def get_model(self):
        return self.search_model.model

    def get_model_field(self):
        return get_related_field(
            model=self.get_model(), related_field_name=self.field
        )

    @property
    def label(self):
        return self._label or self.get_model_field().verbose_name


class SearchModel(AppsModuleLoaderMixin):
    _loader_module_name = 'search'
    _model_search_relationships = {}
    _registry = {}

    @staticmethod
    def flatten_list(value):
        if isinstance(value, (str, bytes)):
            yield value
        else:
            for item in value:
                if isinstance(item, Iterable) and not isinstance(item, (str, bytes)):
                    yield from SearchModel.flatten_list(value=item)
                else:
                    yield item

    @staticmethod
    def function_return_same(value):
        return value

    @staticmethod
    def initialize():
        # Hide a circular import.
        from .handlers import (
            handler_factory_deindex_instance, handler_index_instance
        )

        for search_model in SearchModel.all():
            post_save.connect(
                dispatch_uid='search_handler_index_instance_{}'.format(search_model),
                receiver=handler_index_instance,
                sender=search_model.model
            )
            pre_delete.connect(
                dispatch_uid='search_handler_deindex_instance_{}'.format(search_model),
                receiver=handler_factory_deindex_instance(search_model=search_model),
                sender=search_model.model,
                weak=False
            )
            for proxy in search_model.proxies:
                post_save.connect(
                    dispatch_uid='search_handler_index_instance_{}'.format(search_model),
                    receiver=handler_index_instance,
                    sender=proxy
                )
                pre_delete.connect(
                    dispatch_uid='search_handler_deindex_instance_{}'.format(search_model),
                    receiver=handler_factory_deindex_instance(search_model=search_model),
                    sender=proxy,
                    weak=False
                )

            search_model._initialize()

    @classmethod
    def all(cls):
        return sorted(
            list(set(cls._registry.values())), key=lambda x: x.label
        )

    @classmethod
    def as_choices(cls):
        return cls._registry

    @classmethod
    def get(cls, name):
        try:
            result = cls._registry[name]
        except KeyError:
            raise KeyError(_('No search model matching the query'))
        if not hasattr(result, 'serializer'):
            result.serializer = import_string(dotted_path=result.serializer_path)

        return result

    @classmethod
    def get_default(cls):
        for search_class in cls.all():
            if search_class.default:
                return search_class

    @classmethod
    def get_for_model(cls, instance):
        return cls.get(name=instance._meta.label)

    def __init__(
        self, app_label, model_name, serializer_path, default=False,
        label=None, list_mode=None, permission=None, queryset=None
    ):
        self.default = default
        self._label = label
        self.app_label = app_label
        self.list_mode = list_mode or LIST_MODE_CHOICE_LIST
        self.model_name = model_name
        self._proxies = []  # Lazy
        self.permission = permission
        self.queryset = queryset
        self.search_fields = []
        self.serializer_path = serializer_path

        if default:
            for search_class in self.__class__._registry.values():
                search_class.default = False

        self.__class__._registry[self.get_full_name()] = self

    def __repr__(self):
        return '<{}: {}>'.format(
            self.__class__.__name__, self.label
        )

    def __str__(self):
        return force_text(s=self.label)

    def _initialize(self):
        for search_field in self.search_fields:
            related_model = get_related_field(
                model=self.model, related_field_name=search_field.field
            ).model

            if related_model != self.model:
                self.__class__._model_search_relationships.setdefault(
                    self.model, set()
                )
                self.__class__._model_search_relationships[self.model].add(
                    related_model
                )

                self.__class__._model_search_relationships.setdefault(
                    related_model, set()
                )
                self.__class__._model_search_relationships[related_model].add(
                    self.model
                )

    def add_model_field(self, *args, **kwargs):
        """
        Add a search field that directly belongs to the parent SearchModel.
        """
        search_field = SearchField(self, *args, **kwargs)
        self.search_fields.append(search_field)

    def add_proxy_model(self, app_label, model_name):
        self._proxies.append(
            {
                'app_label': app_label, 'model_name': model_name
            }
        )

        self.__class__._registry['{}.{}'.format(app_label, model_name)] = self

    def get_fields_simple_list(self):
        """
        Returns a list of the fields for the SearchModel.
        """
        result = []
        for search_field in self.search_fields:
            result.append(
                (search_field.get_full_name(), search_field.label)
            )

        return sorted(result, key=lambda x: x[1])

    def get_full_name(self):
        return '{}.{}'.format(self.app_label, self.model_name)

    def get_queryset(self):
        if self.queryset:
            return self.queryset()
        else:
            return self.model.objects.all()

    def get_search_field(self, full_name):
        try:
            return self.search_fields[full_name]
        except KeyError:
            raise KeyError('No search field named: %s' % full_name)

    @cached_property
    def label(self):
        if not self._label:
            self._label = self.model._meta.verbose_name
        return self._label

    @cached_property
    def model(self):
        return apps.get_model(
            app_label=self.app_label, model_name=self.model_name
        )

    @cached_property
    def pk(self):
        return self.get_full_name()

    @cached_property
    def proxies(self):
        result = []
        for proxy in self._proxies:
            result.append(
                apps.get_model(
                    app_label=proxy['app_label'], model_name=proxy['model_name']
                )
            )
        return result

    def sieve(self, field_map, instance):
        """
        Method that receives an instance and a field map dictionary
        consisting of attribute names and transformations to apply.
        Returns a dictionary of the instance values with their respective
        transformations. Makes it easy to pre process an instance before
        indexing it.
        """
        result = {}
        for field in field_map:
            try:
                value = ResolverPipelineModelAttribute.resolve(
                    attribute=field, obj=instance
                )
                try:
                    value = list(SearchModel.flatten_list(value))
                    if value == [None]:
                        value = None
                    else:
                        value = ''.join(value)
                except TypeError:
                    """Value is not a list."""
            except ResolverPipelineError:
                """Not fatal."""
            else:
                result[field] = field_map[field].get(
                    'transformation', SearchModel.function_return_same
                )(value)

        return result



from django.core.exceptions import ValidationError as DjangoValidationError
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse

from mayan.apps.documents.serializers.document_serializers import (
    DocumentSerializer
)
from mayan.apps.documents.serializers.document_type_serializers import (
    DocumentTypeSerializer
)
from mayan.apps.rest_api.serializer_mixins import CreateOnlyFieldSerializerMixin
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .models import DocumentMetadata, DocumentTypeMetadataType, MetadataType
from .permissions import permission_document_metadata_add


class MetadataTypeSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_field': 'pk', 'lookup_url_kwarg': 'metadata_type_id',
                'view_name': 'rest_api:metadatatype-detail'
            },
        }
        fields = (
            'default', 'id', 'label', 'lookup', 'name', 'parser', 'url',
            'validation'
        )
        model = MetadataType


class DocumentTypeMetadataTypeSerializer(serializers.HyperlinkedModelSerializer):
    document_type = DocumentTypeSerializer(read_only=True)
    metadata_type = MetadataTypeSerializer(read_only=True)
    url = serializers.SerializerMethodField()

    class Meta:
        fields = ('document_type', 'id', 'metadata_type', 'required', 'url')
        model = DocumentTypeMetadataType

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:documenttypemetadatatype-detail', kwargs={
                'document_type_id': instance.document_type.pk,
                'metadata_type_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )


class NewDocumentTypeMetadataTypeSerializer(serializers.ModelSerializer):
    metadata_type_id = serializers.IntegerField(
        help_text=_('Primary key of the metadata type to be added.'),
        write_only=True
    )
    url = serializers.SerializerMethodField()

    class Meta:
        fields = (
            'id', 'metadata_type_id', 'required', 'url'
        )
        model = DocumentTypeMetadataType

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:documenttypemetadatatype-detail', kwargs={
                'document_type_id': instance.document_type.pk,
                'metadata_type_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def validate(self, attrs):
        attrs['document_type'] = self.context['document_type']
        attrs['metadata_type'] = MetadataType.objects.get(
            pk=attrs.pop('metadata_type_id')
        )

        instance = DocumentTypeMetadataType(**attrs)
        try:
            instance.full_clean()
        except DjangoValidationError as exception:
            raise ValidationError(exception)

        return attrs


class WritableDocumentTypeMetadataTypeSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField()

    class Meta:
        fields = (
            'id', 'required', 'url'
        )
        model = DocumentTypeMetadataType

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:documenttypemetadatatype-detail', kwargs={
                'document_type_id': instance.document_type.pk,
                'metadata_type_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )


class DocumentMetadataSerializer(
    CreateOnlyFieldSerializerMixin, serializers.HyperlinkedModelSerializer
):
    metadata_type_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the metadata type to be added to the document.'
        ), source_model=MetadataType,
        source_permission=permission_document_metadata_add
    )
    document = DocumentSerializer(read_only=True)
    metadata_type = MetadataTypeSerializer(read_only=True)
    url = serializers.SerializerMethodField()

    class Meta:
        create_only_fields = ('metadata_type_id',)
        fields = (
            'document', 'id', 'metadata_type', 'metadata_type_id', 'url',
            'value'
        )
        model = DocumentMetadata
        read_only_fields = ('document', 'metadata_type',)

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:documentmetadata-detail', kwargs={
                'document_id': instance.document.pk,
                'metadata_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def validate(self, attrs):
        if self.instance:
            self.instance.value = attrs['value']

            try:
                self.instance.full_clean()
            except DjangoValidationError as exception:
                raise ValidationError(exception)

            attrs['value'] = self.instance.value

            return attrs
        else:
            attrs['document'] = self.context['external_object']
            attrs['metadata_type'] = MetadataType.objects.get(
                pk=attrs.pop('metadata_type_id').pk
            )

            instance = DocumentMetadata(**attrs)
            try:
                instance.full_clean()
            except DjangoValidationError as exception:
                raise ValidationError(exception)

            attrs['value'] = instance.value

            return attrs



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import mayan

PACKAGE_NAME = 'mayan-edms'
PACKAGE_DIR = 'mayan'

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


def find_packages(directory):
    # Compile the list of packages available, because distutils doesn't have
    # an easy way to do this.
    packages, data_files = [], []
    root_dir = os.path.dirname(__file__)
    if root_dir != '':
        os.chdir(root_dir)

    for dirpath, dirnames, filenames in os.walk(directory):
        if not dirpath.startswith('mayan/media'):
            # Ignore dirnames that start with '.'
            if os.path.basename(dirpath).startswith('.'):
                continue
            if '__init__.py' in filenames:
                packages.append('.'.join(fullsplit(dirpath)))
            elif filenames:
                data_files.append(
                    [
                        dirpath, [
                            os.path.join(dirpath, filename) for filename in filenames
                        ]
                    ]
                )

    return packages

install_requires = """
django==2.2.24
Pillow==7.1.2
PyPDF2==1.26.0
PyYAML==5.3.1
Whoosh==2.7.4
bleach==3.1.5
celery==4.4.7
django-activity-stream==0.8.0
django-celery-beat==2.0.0
django-colorful==1.3
django-cors-headers==3.2.1
django-formtools==2.2
django-mathfilters==1.0.0
django-model-utils==4.0.0
django-mptt==0.12.0
django-pure-pagination==0.3.0
django-qsstats-magic==1.1.0
django-solo==1.1.5
django-stronghold==0.4.0
django-widget-tweaks==1.4.8
djangorestframework==3.11.2
djangorestframework-recursive==0.1.2
drf-yasg==1.17.1
extract-msg==0.23.3
flanker==0.9.11
flex==6.14.1
furl==2.1.0
fusepy==3.0.1
gevent==20.4.0
graphviz==0.14
gunicorn==20.0.4
mock==4.0.2
node-semver==0.8.0
packaging==20.3
pycountry==20.7.3
pycryptodome==3.9.7
python-dateutil==2.8.1
python-magic==0.4.22
python_gnupg==0.4.6
pytz==2020.1
requests==2.25.1
sh==1.14.1
swagger-spec-validator==2.5.0
whitenoise==5.0.1
""".split()

with open(file='README.rst') as file_object:
    readme = file_object.read()

with open(file='HISTORY.rst') as file_object:
    history = file_object.read()

setup(
    author='Roberto Rosario',
    author_email='roberto.rosario@mayan-edms.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Communications :: File Sharing',
    ],
    description=mayan.__description__,
    include_package_data=True,
    install_requires=install_requires,
    license='Apache 2.0',
    long_description=readme + '\n\n' + history,
    name=PACKAGE_NAME,
    packages=find_packages(PACKAGE_DIR),
    platforms=['any'],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    scripts=['mayan/bin/mayan-edms.py'],
    url='https://gitlab.com/mayan-edms/mayan-edms',
    version=mayan.__version__,
    zip_safe=False,
)



from mayan.apps.rest_api import generics

from .models import Announcement
from .permissions import (
    permission_announcement_create, permission_announcement_delete,
    permission_announcement_edit, permission_announcement_view
)
from .serializers import AnnouncementSerializer


class APIAnnouncementListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the announcements.
    post: Create a new announcement.
    """
    mayan_object_permissions = {'GET': (permission_announcement_view,)}
    mayan_view_permissions = {'POST': (permission_announcement_create,)}
    ordering_fields = (
        'enabled', 'end_datetime', 'id', 'label', 'start_datetime'
    )
    queryset = Announcement.objects.all()
    serializer_class = AnnouncementSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIAnnouncementView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected announcement.
    get: Return the details of the selected announcement.
    patch: Edit the selected announcement.
    put: Edit the selected announcement.
    """
    lookup_url_kwarg = 'announcement_id'
    mayan_object_permissions = {
        'DELETE': (permission_announcement_delete,),
        'GET': (permission_announcement_view,),
        'PATCH': (permission_announcement_edit,),
        'PUT': (permission_announcement_edit,)
    }
    queryset = Announcement.objects.all()
    serializer_class = AnnouncementSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }



import logging

from django.template import Context, Library, Variable, VariableDoesNotExist
from django.template.defaultfilters import truncatechars
from django.template.loader import get_template
from django.utils.translation import ugettext_lazy as _

from mayan.apps.appearance.settings import setting_max_title_length

from ..icons import icon_list_mode_items, icon_list_mode_list

logger = logging.getLogger(name=__name__)
register = Library()


@register.simple_tag(takes_context=True)
def views_calculate_title(context):
    title = ''
    title_full = ''

    if context.get('title'):
        title_full = context.get('title')
        title = truncatechars(
            title_full, arg=setting_max_title_length.value
        )
    else:
        if context.get('delete_view'):
            title = _('Confirm delete')
            title_full = title
        else:
            if context.get('form'):
                if context.get('object'):
                    title = _('Edit %s') % context.get('object')
                    title_full = title
                else:
                    title = _('Confirm')
                    title_full = title
            else:
                if context.get('read_only'):
                    title = _('Details for: %s') % context.get('object')
                    title_full = title
                else:
                    if context.get('object'):
                        title = _('Edit: %s') % context.get('object')
                        title_full = title
                    else:
                        if context.get('create_view') or context.get('form'):
                            title = _('Create')
                            title_full = title

    return {'title': title, 'title_full': title_full}


@register.simple_tag(takes_context=True)
def views_get_list_mode_icon(context):
    if context.get('list_as_items', False):
        return icon_list_mode_list
    else:
        return icon_list_mode_items


@register.simple_tag(takes_context=True)
def views_get_list_mode_querystring(context):
    try:
        request = context.request
    except AttributeError:
        # Simple request extraction failed. Might not be a view context.
        # Try alternate method.
        try:
            request = Variable('request').resolve(context)
        except VariableDoesNotExist:
            # There is no request variable, most probable a 500 in a test
            # view. Don't return any resolved request.
            logger.warning('No request variable, aborting request resolution')
            return ''

    # We do this to get an mutable copy we can modify
    querystring = request.GET.copy()

    list_as_items = context.get('list_as_items', False)

    if list_as_items:
        querystring['_list_mode'] = 'list'
    else:
        querystring['_list_mode'] = 'items'

    return '?{}'.format(querystring.urlencode())


@register.simple_tag(takes_context=True)
def views_render_subtemplate(context, template_name, template_context):
    """
    Renders the specified template with the mixed parent and
    subtemplate contexts.
    """
    new_context = Context(context.flatten())
    new_context.update(Context(template_context))
    return get_template(template_name).render(new_context.flatten())



from ..classes import ModelCopy


class CommonAPITestMixin:
    def _request_content_type_list_api_view(self):
        return self.get(viewname='rest_api:content-type-list')


class CommonViewTestMixin:
    def _request_about_view(self):
        return self.get(viewname='common:about_view')


class ObjectCopyTestMixin:
    _test_copy_method = None

    def test_copy_method(self, exclude_fields=None, test_object=None, test_object_copy=None):
        exclude_fields = exclude_fields or ()

        if not test_object:
            test_object = self.test_object

        model_copy = ModelCopy.get(model=test_object._meta.model)
        if not test_object_copy:
            test_object_copy = test_object.copy_instance()

        if test_object == test_object_copy:
            return

        if self._test_copy_method:
            test_objects = zip(
                getattr(test_object, self._test_copy_method)(),
                getattr(test_object_copy, self._test_copy_method)()
            )
        else:
            test_objects = ((test_object, test_object_copy),)

        for test_object, test_object_copy in test_objects:
            for field in model_copy.fields:
                if field not in exclude_fields:
                    if field in model_copy.fields_reverse_related:
                        related_test_objects = zip(
                            getattr(test_object, field).all(),
                            getattr(test_object_copy, field).all()
                        )
                        exclude_fields = exclude_fields + (
                            test_object._meta.get_field(field).remote_field.name,
                        )

                        for related_test_object, related_test_object_copy in related_test_objects:
                            self.test_copy_method(
                                exclude_fields=exclude_fields,
                                test_object=related_test_object,
                                test_object_copy=related_test_object_copy
                            )
                    elif field in model_copy.fields_foreign_keys:
                        related_test_objects = (
                            (
                                getattr(test_object, field),
                                getattr(test_object_copy, field)
                            ),
                        )
                        exclude_fields = exclude_fields + (
                            test_object._meta.get_field(field).remote_field.name,
                        )

                        for related_test_object, related_test_object_copy in related_test_objects:
                            self.test_copy_method(
                                exclude_fields=exclude_fields,
                                test_object=related_test_object,
                                test_object_copy=related_test_object_copy
                            )
                    elif field in model_copy.fields_related_one_to_one:
                        related_test_objects = (
                            (
                                getattr(test_object, field),
                                getattr(test_object_copy, field)
                            ),
                        )
                        exclude_fields = exclude_fields + (
                            test_object._meta.get_field(field).remote_field.name,
                        )

                        for related_test_object, related_test_object_copy in related_test_objects:
                            self.test_copy_method(
                                exclude_fields=exclude_fields,
                                test_object=related_test_object,
                                test_object_copy=related_test_object_copy
                            )
                    else:
                        original_value = test_object._meta.default_manager.filter(
                            pk=test_object.pk
                        ).values_list(field, flat=True).first()

                        try:
                            conditional = model_copy.unique_conditional[field]
                        except KeyError:
                            """This field has no conditional entry"""
                        else:
                            new_instance_dictionary = test_object_copy.__dict__.copy()
                            if new_instance_dictionary[field].endswith('_1'):
                                new_instance_dictionary[field] = new_instance_dictionary[field][:-2]
                                if conditional(
                                    instance=test_object, new_instance_dictionary=new_instance_dictionary
                                ):
                                    original_value = '{}_1'.format(original_value)

                        if field in model_copy.fields_unique:
                            original_value = '{}_1'.format(original_value)

                        copy_value = test_object_copy._meta.default_manager.filter(
                            pk=test_object_copy.pk
                        ).values_list(field, flat=True).first()

                        self.assertEqual(original_value, copy_value)


class ObjectCopyViewTestMixin:
    def _request_object_copy_view(self):
        return self.post(
            kwargs=self.test_object_view_kwargs, viewname='common:object_copy'
        )



import logging
import os

from django.core.files import File
from django.db import models

from mayan.apps.django_gpg.exceptions import DecryptionError
from mayan.apps.django_gpg.models import Key
from mayan.apps.documents.models import DocumentFile
from mayan.apps.storage.utils import NamedTemporaryFile, mkstemp

from .events import (
    event_detached_signature_created,
    event_embedded_signature_created
)


logger = logging.getLogger(name=__name__)


class DetachedSignatureManager(models.Manager):
    def sign_document_file(
        self, document_file, key, passphrase=None, user=None
    ):
        with NamedTemporaryFile() as temporary_file_object:
            with document_file.open() as file_object:
                key.sign_file(
                    binary=True, detached=True, file_object=file_object,
                    output=temporary_file_object.name,
                    passphrase=passphrase
                )
            temporary_file_object.seek(0)

            instance = self.model(
                document_file=document_file,
                signature_file=File(file=temporary_file_object)
            )
            instance._event_ignore = True
            instance.save()
            event_detached_signature_created.commit(
                action_object=instance, actor=user, target=document_file
            )
            return instance


class EmbeddedSignatureManager(models.Manager):
    def open_signed(self, document_file, file_object):
        for signature in self.filter(document_file=document_file):
            try:
                return self.open_signed(
                    file_object=Key.objects.decrypt_file(
                        file_object=file_object
                    ), document_file=document_file
                )
            except DecryptionError:
                file_object.seek(0)
                return file_object
        else:
            return file_object

    def sign_document_file(
        self, document_file, key, passphrase=None, user=None
    ):
        temporary_file_object, temporary_filename = mkstemp()

        try:
            with document_file.open() as file_object:
                key.sign_file(
                    binary=True, file_object=file_object,
                    output=temporary_filename, passphrase=passphrase
                )
        except Exception:
            raise
        else:
            # The result of key.sign_file does not contain the signtarure ID.
            # Verify the signed file to obtain the signature ID.
            with open(file=temporary_filename, mode='rb') as file_object:
                result = Key.objects.verify_file(
                    file_object=file_object
                )

            with open(file=temporary_filename, mode='rb') as file_object:
                document_file.document.file_new(
                    file_object=file_object, _user=user
                )
            instance = self.get(signature_id=result.signature_id)
            event_embedded_signature_created.commit(
                action_object=instance, actor=user, target=document_file
            )
            return instance
        finally:
            os.unlink(temporary_filename)

    def unsigned_document_files(self):
        return DocumentFile.objects.exclude(
            pk__in=self.values('document_file')
        )
